@@ -6,7 +6,7 @@ from future.builtins import object
 
 from itertools import islice, chain
 from collections import namedtuple
-from copy import deepcopy
+from copy import copy, deepcopy
 
 from limpyd.model import RedisModel
 from limpyd.collection import CollectionManager, ParsedFilter
@@ -14,6 +14,7 @@ from limpyd.fields import (SetField, ListField, SortedSetField, MultiValuesField
                            RedisField, SingleValueField)
 from limpyd.exceptions import DoesNotExist
 from limpyd.contrib.database import PipelineDatabase
+from limpyd.utils import make_key
 
 SORTED_SCORE = 'sorted_score'
 DEFAULT_STORE_TTL = 60
@@ -39,8 +40,8 @@ class ExtendedCollectionManager(CollectionManager):
         },
     }
 
-    def __init__(self, cls):
-        super(ExtendedCollectionManager, self).__init__(cls)
+    def __init__(self, model):
+        super(ExtendedCollectionManager, self).__init__(model)
 
         self._lazy_collection['intersects'] = set()
 
@@ -53,54 +54,50 @@ class ExtendedCollectionManager(CollectionManager):
 
         self._values = None  # Will store parameters used to retrieve values
 
+    def clone(self):
+        new = super(ExtendedCollectionManager, self).clone()
+        new._has_sortedsets = self._has_sortedsets
+        new._sort_by_sortedset = self._sort_by_sortedset.copy() if self._sort_by_sortedset is not None else None
+        new._store = self._store
+        new.stored_key = self.stored_key
+        new._stored_len = self._stored_len
+        new._values = {key: copy(value) for key, value in self._values.items()} if self._values is not None else None
+        return new
+
     def _list_to_set(self, list_key, set_key):
         """
         Store all content of the given ListField in a redis set.
-        Use scripting if available to avoid retrieving all values locally from
-        the list before sending them back to the set
+        Use lua scripting to avoid retrieving all values locally from the list before sending them
+        back to the set
         """
-        if self.cls.database.support_scripting():
-            self.cls.database.call_script(
-                # be sure to use the script dict at the class level
-                # to avoid registering it many times
-                script_dict=self.__class__.scripts['list_to_set'],
-                keys=[list_key, set_key]
-            )
-        else:
-            conn = self.cls.get_connection()
-            conn.sadd(set_key, *conn.lrange(list_key, 0, -1))
+        self.model.database.call_script(
+            # be sure to use the script dict at the class level
+            # to avoid registering it many times
+            script_dict=self.__class__.scripts['list_to_set'],
+            keys=[list_key, set_key]
+        )
 
-    @property
-    def _collection(self):
+    def _fetch_collection(self, apply_slice=None):
         """
         Effectively retrieve data according to lazy_collection.
         If we have a stored collection, without any result, return an empty list
         """
-        old_sort_limits_and_len_mode = None if self._sort_limits is None else self._sort_limits.copy(), self._len_mode
-        old_sorts = None if self._sort is None else self._sort.copy(),\
-                    None if self._sort_by_sortedset is None else self._sort_by_sortedset.copy()
-        try:
-            if self.stored_key and not self._stored_len:
-                if self._len_mode:
-                    self._len = 0
-                    self._len_mode = False
-                self._sort_limits = {}
-                return []
+        if self.stored_key and not self._stored_len:
+            self._sort_limits = {}
+            self._cache_empty_collection()
+            return
 
-            # Manage sort desc added by original `__getitem__` when we sort by score
-            if self._sort_by_sortedset and self._sort and self._sort.get('desc'):
-                self._sort = None
-                self._sort_by_sortedset['desc'] = not self._sort_by_sortedset.get('desc', False)
+        # Manage sort desc added by original `__getitem__` when we sort by score
+        if self._sort_by_sortedset and self._sort and self._sort.get('desc'):
+            self._sort = None
+            self._sort_by_sortedset['desc'] = not self._sort_by_sortedset.get('desc', False)
 
-            return super(ExtendedCollectionManager, self)._collection
-        finally:
-            self._sort_limits, self._len_mode = old_sort_limits_and_len_mode
-            self._sort, self._sort_by_sortedset = old_sorts
+        super(ExtendedCollectionManager, self)._fetch_collection(apply_slice=apply_slice)
 
     def _prepare_sets(self, sets):
         """
-        The original "_prepare_sets" method simple return the list of sets in
-        _lazy_collection, know to be all keys of redis sets.
+        The original "_prepare_sets" method simply return the list of sets in
+        `_lazy_collection`, which only contains redis sets keys.
         As the new "intersect" method can accept different types of "set", we
         have to handle them because we must return only keys of redis sets.
         """
@@ -109,7 +106,7 @@ class ExtendedCollectionManager(CollectionManager):
             raise DoesNotExist('This collection is based on a previous one, '
                                'stored at a key that does not exist anymore.')
 
-        conn = self.cls.get_connection()
+        conn = self.connection
 
         all_sets = set()
         tmp_keys = set()
@@ -137,32 +134,25 @@ class ExtendedCollectionManager(CollectionManager):
             if is_tmp:
                 tmp_keys.add(key)
 
+        prepared_sets = []
         for set_ in sets:
+            if isinstance(set_, ParsedFilter):
+                # We have a RedisModel and we'll use its pk, or a RedisField
+                # (single value) and we'll use its value
+                if isinstance(set_.value, RedisModel):
+                    set_ = ParsedFilter(set_.index, set_.suffix, set_.extra_field_parts, set_.value.pk.get(), set_.related_filters)
+                elif isinstance(set_.value, SingleValueField):
+                    set_ = ParsedFilter(set_.index, set_.suffix, set_.extra_field_parts, set_.value.proxy_get(), set_.related_filters)
+                elif isinstance(set_.value, RedisField):
+                    raise ValueError(u'Invalid filter value for %s: %s' % (set_.index.field.name, set_.value))
+            prepared_sets.append(set_)
+
+        for set_ in self._reduce_related_filters(prepared_sets):
             if isinstance(set_, str):
                 add_key(set_)
             elif isinstance(set_, ParsedFilter):
-
-                value = set_.value
-                # We have a RedisModel and we'll use its pk, or a RedisField
-                # (single value) and we'll use its value
-                if isinstance(value, RedisModel):
-                    value = value.pk.get()
-                elif isinstance(value, SingleValueField):
-                    value = value.proxy_get()
-                elif isinstance(value, RedisField):
-                    raise ValueError(u'Invalid filter value for %s: %s' % (set_.index.field.name, value))
-
-                for index_key, key_type, is_tmp in set_.index.get_filtered_keys(
-                            set_.suffix,
-                            accepted_key_types=self._accepted_key_types,
-                            *(set_.extra_field_parts + [value])
-                        ):
-                    if key_type not in self._accepted_key_types:
-                        raise ValueError('The index key returned by the index %s is not valid' % (
-                            set_.index.__class__.__name__
-                        ))
+                for index_key, key_type, is_tmp in self._prepare_parsed_filter(set_):
                     add_key(index_key, key_type, is_tmp)
-
             elif isinstance(set_, SetField):
                 # Use the set key. If we need to intersect, we'll use
                 # sunionstore, and if not, store accepts set
@@ -175,7 +165,7 @@ class ExtendedCollectionManager(CollectionManager):
                 add_key(set_.key, 'list')
             elif isinstance(set_, tuple) and len(set_):
                 # if we got a list or set, create a redis set to hold its values
-                tmp_key = self._unique_key()
+                tmp_key = self._unique_key('tmp')
                 conn.sadd(tmp_key, *set_)
                 add_key(tmp_key, 'set', True)
             else:
@@ -189,7 +179,7 @@ class ExtendedCollectionManager(CollectionManager):
                 # we have many sets/lists, we need to convert them to sets
                 for list_key in lists:
                     # many sets, convert the list to a simple redis set
-                    tmp_key = self._unique_key()
+                    tmp_key = self._unique_key('tmp')
                     self._list_to_set(list_key, tmp_key)
                     add_key(tmp_key, 'set', True)
 
@@ -199,9 +189,9 @@ class ExtendedCollectionManager(CollectionManager):
         """
         Add more filters to the collection
         """
-        return self._add_filters(**filters)
+        return self.clone()._add_filters(**filters)
 
-    def intersect(self, *sets):
+    def _apply_intersect(self, *sets):
         """
         Add a list of sets to the existing list of sets to check. Returns self
         for chaining.
@@ -237,6 +227,9 @@ class ExtendedCollectionManager(CollectionManager):
         self._lazy_collection['intersects'].update(sets_)
         return self
 
+    def intersect(self, *sets):
+        return self.clone()._apply_intersect(*sets)
+
     def _combine_sets(self, sets, final_set):
         """
         Given a list of set, combine them to create the final set that will be
@@ -244,7 +237,7 @@ class ExtendedCollectionManager(CollectionManager):
         If we have a least a sorted set, use zinterstore insted of sunionstore
         """
         if self._has_sortedsets:
-            self.cls.get_connection().zinterstore(final_set, list(sets))
+            self.connection.zinterstore(final_set, list(sets))
         else:
             final_set = super(ExtendedCollectionManager, self)._combine_sets(sets, final_set)
         return final_set
@@ -257,7 +250,7 @@ class ExtendedCollectionManager(CollectionManager):
         options, call zrange on the final set wich is the result of a call to
         zinterstore.
         """
-        conn = self.cls.get_connection()
+        conn = self.connection
 
         # we have a sorted set without need to sort, use zrange
         if self._has_sortedsets and sort_options is None:
@@ -272,16 +265,30 @@ class ExtendedCollectionManager(CollectionManager):
 
             return conn.lrange(final_set, 0, -1)
 
-        # normal call
-        return super(ExtendedCollectionManager, self)._final_redis_call(
-                                                        final_set, sort_options)
+        keys_to_delete_after = set()
+        try:
+            if final_set and self._sort_by_sortedset_before:
+                # TODO: if we have filters, maybe apply _zet_to_keys to only
+                #       intersected values
+                base_tmp_key, tmp_keys = self._prepare_sort_by_score(None, sort_options)
+                # new keys have to be deleted once the final sort is done
+                keys_to_delete_after.add(base_tmp_key)
+                keys_to_delete_after |= tmp_keys
+
+            # normal call
+            return super(ExtendedCollectionManager, self)._final_redis_call(
+                                                            final_set, sort_options)
+
+        finally:
+            if keys_to_delete_after:
+                conn.delete(*keys_to_delete_after)
 
     def _collection_length(self, final_set):
         """
         Return the length of the final collection, directly asking redis for the
         count without calling sort
         """
-        conn = self.cls.get_connection()
+        conn = self.connection
 
         # we have a sorted set without need to sort, use zcard
         if self._has_sortedsets:
@@ -296,7 +303,7 @@ class ExtendedCollectionManager(CollectionManager):
         # normal call
         return super(ExtendedCollectionManager, self)._collection_length(final_set)
 
-    def sort(self, **parameters):
+    def _apply_sort(self, **parameters):
         """
         Enhance the default sort method to accept a new parameter "by_score", to
         use instead of "by" if you want to sort by the score of a sorted set.
@@ -330,7 +337,7 @@ class ExtendedCollectionManager(CollectionManager):
             if by and isinstance(by, RedisField):
                 parameters['by'] = by.name
 
-        super(ExtendedCollectionManager, self).sort(**parameters)
+        super(ExtendedCollectionManager, self)._apply_sort(**parameters)
 
         if is_sortedset:
             self._sort_by_sortedset = self._sort
@@ -349,7 +356,7 @@ class ExtendedCollectionManager(CollectionManager):
         If a value in values is not on the sorted set, it's still saved as a key
         but with a default value ('' is alpha is True, else '-inf')
         """
-        conn = self.cls.get_connection()
+        conn = self.connection
         default = '' if alpha else '-inf'
         if values is None:
             # no values given, we get scores from the whole sorted set
@@ -358,10 +365,10 @@ class ExtendedCollectionManager(CollectionManager):
         else:
             # we have values, we'll get only their scores
 
-            if isinstance(self.cls.database, PipelineDatabase):
+            if isinstance(self.model.database, PipelineDatabase):
                 # if available, use the pipeline of our database to get all
                 # scores in one redis call
-                with self.cls.database.pipeline(transaction=False) as pipe:
+                with self.model.database.pipeline(transaction=False) as pipe:
                     for value in values:
                         pipe.zscore(key, value)
                     scores = pipe.execute()
@@ -380,14 +387,14 @@ class ExtendedCollectionManager(CollectionManager):
                 result.append((value, score))
 
         # create a temporary key for each (value,score) tuple
-        base_tmp_key = self._unique_key()
+        base_tmp_key = self._unique_key('tmp')
         conn.set(base_tmp_key, 'working...')  # only to "reserve" the main tmp key
-        tmp_keys = []
+        tmp_keys = set()
         # use a mapping dict (tmp_key_with_value=>score) to use in mset
         mapping = {}
         for value, score in result:
-            tmp_key = '%s:%s' % (base_tmp_key, value)
-            tmp_keys.append(tmp_key)
+            tmp_key = make_key(base_tmp_key, value)
+            tmp_keys.add(tmp_key)
             mapping[tmp_key] = score
         # set all keys in one call
         conn.mset(mapping)
@@ -423,7 +430,7 @@ class ExtendedCollectionManager(CollectionManager):
 
         return base_tmp_key, tmp_keys
 
-    def _prepare_results(self, results):
+    def _prepare_results(self, results, apply_slice=None):
         """
         Sort results by score if not done before (faster, if we have no values to
         retrieve, or slice)
@@ -432,7 +439,7 @@ class ExtendedCollectionManager(CollectionManager):
         # (no slice), we can do it know, by creating keys for each values with
         # the sorted set score, and sort on them
         if self._sort_by_sortedset_after and (len(results) > 1 or self._values):
-            conn = self.cls.get_connection()
+            conn = self.connection
 
             sort_params = {}
             base_tmp_key, tmp_keys = self._prepare_sort_by_score(results, sort_params)
@@ -442,31 +449,36 @@ class ExtendedCollectionManager(CollectionManager):
             conn.sadd(final_set, *results)
 
             # apply the sort
-            results = conn.sort(final_set, **sort_params)
+            results = list(conn.sort(final_set, **sort_params))
 
             # finally delete all temporary keys
-            conn.delete(*(tmp_keys + [final_set, base_tmp_key]))
+            conn.delete(*(tmp_keys | {final_set, base_tmp_key}))
 
         if self._store:
             # if store, redis doesn't return result, so don't return anything here
-            return
+            _len_hint = 0
+            results = []
 
-        if self._values and self._values['mode'] != 'flat':
-            results = self._to_values(results)
+        else:
+            results = list(results)
 
-        return super(ExtendedCollectionManager, self)._prepare_results(results)
+            if self._values and self._values['mode'] != 'flat':
+                # regroup results by tuples when we have many values by entry
+                results = list(zip(*([iter(results)] * len(self._values['fields']['names']))))
 
-    def _to_values(self, collection):
-        """
-        Regroup values in tuples or dicts for each "instance".
-        Exemple: Given this result from redis: ['id1', 'name1', 'id2', 'name2']
-         tuples: [('id1', 'name1'), ('id2', 'name2')]
-         dicts:  [{'id': 'id1', 'name': 'name1'}, {'id': 'id2', 'name': 'name2'}]
-        """
-        result = zip(*([iter(collection)] * len(self._values['fields']['names'])))
-        if self._values['mode'] == 'dicts':
-            result = (dict(zip(self._values['fields']['names'], a_result)) for a_result in result)
-        return result
+            if apply_slice is not None:
+                results = results[apply_slice]
+            _len_hint = len(results)
+
+        results, iterator_function = super(ExtendedCollectionManager, self)._prepare_results(results, _len_hint)
+
+        if self._values and self._values['mode'] == 'dicts':
+            iterator_function = self._to_values_dict
+
+        return results, iterator_function
+
+    def _to_values_dict(self, collection_entry):
+        return dict(zip(self._values['fields']['names'], collection_entry))
 
     @property
     def _sort_by_sortedset_before(self):
@@ -537,26 +549,9 @@ class ExtendedCollectionManager(CollectionManager):
             sets = sets[::]
             sets.extend(self._lazy_collection['intersects'])
             if not self._lazy_collection['sets'] and not self.stored_key:
-                sets.append(self.cls.get_field('pk').collection_key)
+                sets.append(self.model.get_field('pk').collection_key)
 
-        final_set, keys_to_delete_later = super(ExtendedCollectionManager,
-                                    self)._get_final_set(sets, pk, sort_options)
-
-        # if we have a slice and we want to sort by the score of a sorted set,
-        # as redis sort command doesn't handle this, we have to create keys for
-        # each value of the sorted set and sort on them
-        # @antirez, y u don't allow this !!??!!
-        if final_set and self._sort_by_sortedset_before:
-            # TODO: if we have filters, maybe apply _zet_to_keys to only
-            #       intersected values
-            base_tmp_key, tmp_keys = self._prepare_sort_by_score(None, sort_options)
-            # new keys have to be deleted once the final sort is done
-            if not keys_to_delete_later:
-                keys_to_delete_later = []
-            keys_to_delete_later.append(base_tmp_key)
-            keys_to_delete_later += tmp_keys
-
-        return final_set, keys_to_delete_later
+        return super(ExtendedCollectionManager, self)._get_final_set(sets, pk, sort_options)
 
     def _add_filters(self, **filters):
         """
@@ -593,7 +588,7 @@ class ExtendedCollectionManager(CollectionManager):
                 else:
                     # create an ParsedFilter which will be used in _prepare_sets
                     index, suffix, extra_field_parts = self._parse_filter_key(key)
-                    parsed_filter = ParsedFilter(index, suffix, extra_field_parts, value)
+                    parsed_filter = ParsedFilter(index, suffix, extra_field_parts, value, None)
                     self._lazy_collection['sets'].append(parsed_filter)
 
                 string_filters.pop(key)
@@ -648,10 +643,10 @@ class ExtendedCollectionManager(CollectionManager):
                 final_fields['names'].append(field_name)
                 final_fields['keys'].append('#')
             else:
-                if not self.cls.has_field(field_name):
+                if not self.model.has_field(field_name):
                     raise ValueError("%s if not a valid field to get from collection"
-                                     " for %s" % (field_name, self.cls.__name__))
-                field = self.cls.get_field(field_name)
+                                     " for %s" % (field_name, self.model.__name__))
+                field = self.model.get_field(field_name)
                 if isinstance(field, MultiValuesField):
                     raise ValueError("It's not possible to get a MultiValuesField"
                                      " from a collection (asked: %s" % field_name)
@@ -676,93 +671,85 @@ class ExtendedCollectionManager(CollectionManager):
         DEFAULT_STORE_TTL, which is 60 secondes. You can pass None if you don't
         want expiration.
         """
-        old_sort_limits_and_len_mode = None if self._sort_limits is None else self._sort_limits.copy(), self._len_mode
-        try:
-            self._store = True
+        clone = self.clone()
+        clone._store = True
 
-            # save sort and values options
-            sort_options = None
-            if self._sort is not None:
-                sort_options = self._sort.copy()
-            values = None
-            if self._values is not None:
-                values = self._values
-                self._values = None
+        # save sort and values options
+        sort_options = None
+        if clone._sort is not None:
+            sort_options = clone._sort.copy()
+        values = None
+        if clone._values is not None:
+            values = clone._values
+            clone._values = None
 
-            # create a key for storage
-            store_key = key or self._unique_key()
-            if self._sort is None:
-                self._sort = {}
-            self._sort['store'] = store_key
+        # create a key for storage
+        store_key = key or clone._unique_key('store')
+        if clone._sort is None:
+            clone._sort = {}
+        clone._sort['store'] = store_key
 
-            # if filter by pk, but without need to get "values", no redis call is done
-            # so force values to get a call to sort (to store result)
-            if self._lazy_collection['pks'] and not self._values:
-                self.values('pk')
+        # if filter by pk, but without need to get "values", no redis call is done
+        # so force values to get a call to sort (to store result)
+        if clone._lazy_collection['pks'] and not clone._values:
+           clone._apply_values('pk')
 
-            # call the collection
-            self._len_mode = False
-            self._collection
+        # call the collection
+        clone._len_mode = False
+        clone._fetch_collection()
 
-            # restore sort and values options
-            self._store = False
-            self._sort = sort_options
-            self._values = values
+        # create the new collection
+        stored_collection = clone.__class__(clone.model).from_stored(store_key)
 
-            # create the new collection
-            stored_collection = self.__class__(self.cls)
-            stored_collection.from_stored(store_key)
+        # apply ttl if needed
+        if ttl is not None:
+            clone.model.get_connection().expire(store_key, ttl)
 
-            # apply ttl if needed
-            if ttl is not None:
-                self.cls.get_connection().expire(store_key, ttl)
+        # set choices about instances/values from the current to the new collection
+        for attr in ('_instances', '_lazy_instances', '_values'):
+            setattr(stored_collection, attr, deepcopy(getattr(clone, attr)))
 
-            # set choices about instances/values from the current to the new collection
-            for attr in ('_instances', '_instances_skip_exist_test', '_values'):
-                setattr(stored_collection, attr, deepcopy(getattr(self, attr)))
-
-            # finally return the new collection
-            return stored_collection
-
-        finally:
-            self._sort_limits, self._len_mode = old_sort_limits_and_len_mode
+        # finally return the new collection
+        return stored_collection
 
     def from_stored(self, key):
         """
         Set the current collection as based on a stored one. The key argument
         is the key off the stored collection.
         """
+        clone = self.clone()
+
         # only one stored key allowed
-        if self.stored_key:
+        if clone.stored_key:
             raise ValueError('This collection is already based on a stored one')
 
         # prepare the collection
-        self.stored_key = key
-        self.intersect(_StoredCollection(self.cls.get_connection(), key))
-        self.sort(by='nosort')  # keep stored order
+        clone.stored_key = key
+        clone._apply_intersect(_StoredCollection(clone.model.get_connection(), key))
+        clone._apply_sort(by='nosort')  # keep stored order
 
         # count the number of results to manage empty result (to not behave like
         # expired key)
-        self._stored_len = self.cls.get_connection().llen(key)
+        clone._stored_len = clone.model.get_connection().llen(key)
 
-        return self
+        return clone
 
     def stored_key_exists(self):
         """
         Check the existence of the stored key (useful if the collection is based
         on a stored one, to check if the redis key still exists)
         """
-        return self.cls.get_connection().exists(self.stored_key)
+        return self.connection.exists(self.stored_key)
 
-    def reset_result_type(self):
+    def _reset_result_type(self):
         """
         Reset the type of values attened for the collection (ie cancel a
         previous "instances" or "values" call)
         """
         self._values = None
-        return super(ExtendedCollectionManager, self).reset_result_type()
+        return super(ExtendedCollectionManager, self)._reset_result_type()
 
-    def values(self, *fields):
+    def _apply_values(self, *fields):
         """
         Ask the collection to return a list of dict of given fields for each
         instance found in the collection.
@@ -777,7 +764,10 @@ class ExtendedCollectionManager(CollectionManager):
         self._values = {'fields': fields, 'mode': 'dicts'}
         return self
 
-    def values_list(self, *fields, **kwargs):
+    def values(self, *fields):
+        return self.clone()._apply_values(*fields)
+
+    def _apply_values_list(self, *fields, **kwargs):
         """
         Ask the collection to return a list of tuples of given fields (in the
         given order) for each instance found in the collection.
@@ -802,21 +792,16 @@ class ExtendedCollectionManager(CollectionManager):
         self._values = {'fields': fields, 'mode': 'flat' if flat else 'tuples'}
         return self
 
+    def values_list(self, *fields, **kwargs):
+        return self.clone()._apply_values_list(*fields, **kwargs)
+
 
 class _StoredCollection(object):
     """
     Simple object to store the key of a stored collection, to be used in
     ExtendedCollectionManager based on a stored collection.
-    The stored key is a list, so it's managed as a ListField (but we only need
-    its key, and lmembers if no scripting)
+    The stored key is a list, so it's managed as a ListField (but we only need its key)
     """
     def __init__(self, connection, key):
         self.connection = connection
         self.key = key
-
-    def lmembers(self):
-        """
-        Return the list of all members of the list, used by _list_to_set if
-        no scripting
-        """
-        return self.connection.lrange(self.key, 0, -1)
