@@ -9,11 +9,10 @@ from inspect import isclass
 from logging import getLogger
 from copy import copy
 
-from redis import VERSION as redispy_version
 from redis.exceptions import RedisError
 
 from limpyd.database import Lock
-from limpyd.utils import cached_property, make_key, normalize
+from limpyd.utils import cached_property, make_key, normalize, NotProvided
 from limpyd.exceptions import *
 
 log = getLogger(__name__)
@@ -181,6 +180,9 @@ class RedisField(RedisProxyCommand):
     _field_parts = 1
     default_indexes = None
 
+    available_getters = {'expire', 'expireat', 'pexpire', 'pexpireat', 'ttl', 'pttl', 'persist'}
+    available_modifiers = set()
+
     def __init__(self, *args, **kwargs):
         """
         Manage all field attributes
@@ -198,7 +200,7 @@ class RedisField(RedisProxyCommand):
                 raise ImplementationError('Cannot set "default" and "unique" together!')
             self.indexable = True
 
-        self.index_classes = kwargs.get('indexes', [])
+        self.index_classes = kwargs.get('indexes') or []
         if self.index_classes:
             if not self.indexable:
                 raise ImplementationError('Cannot pass indexes if not indexable')
@@ -207,6 +209,16 @@ class RedisField(RedisProxyCommand):
         self._creation_order = RedisField._creation_order
         RedisField._creation_order += 1
 
+    def __repr__(self):
+        try:
+            return u'%s (model=%s, name=%s)>' % (
+                super(RedisField, self).__repr__()[:-2],
+                self._model.__name__,
+                self.name,
+            )
+        except Exception:
+            return super(RedisField, self).__repr__()
+
     def get_default_indexes(self):
         if self.__class__.default_indexes is not None:
             return self.__class__.default_indexes
@@ -376,36 +388,54 @@ class RedisField(RedisProxyCommand):
             If no index classes available for this field
 
         """
+
+        if self.attached_to_instance:
+            return self._model.get_field(self.name)._indexes
+
         if not self.indexable:
             return []
-        if not self.index_classes:
-            self.index_classes = self.get_default_indexes()[::1]
         if not self.index_classes:
             raise ImplementationError('%s field is indexable but has no indexes attached' %
                                       self.__class__.__name__)
 
         return [index_class(field=self) for index_class in self.index_classes]
 
-    def has_index(self, index):
-        """Tells if the field have an index matching the current one
+    def get_index(self, index_class=NotProvided, key=NotProvided, prefix=NotProvided):
+        """Get an index matching the given filters.
 
         Parameters
-        -----------
-        index: type or BaseIndex
-            It could be an index instance, or an index class
+        ----------
+        index_class : Optional[Type[BaseIndex]]
+            The index class to filter on
+        key : Optional[str]
+            The key of the index to filter on.
+        prefix : Optional[str]
+            The prefix of the index to filter on
+
+        Raises
+        ------
+        ValueError
+            - If no indexes match the filters
+            - If more than one index match the filters
 
         Returns
         -------
-        bool
-            Will be ``True`` if the current field has an index that is an instance
-            of the given index or of the class of the given index
+        BaseIndex
+            The only index matching all the filter
 
         """
-        klass = index if isclass(index) else index.__class__
-        for one_index in self._indexes:
-            if isinstance(one_index, klass):
-                return True
-        return False
+        indexes = self._indexes
+        if indexes and index_class is not NotProvided:
+            indexes = [index for index in indexes if isinstance(index, index_class)]
+        if indexes and key is not NotProvided:
+            indexes = [index for index in indexes if index.key == key]
+        if indexes and prefix is not NotProvided:
+            indexes = [index for index in indexes if index.prefix == prefix]
+        if len(indexes) > 1:
+            raise ValueError('More than one index matching filters')
+        if not indexes:
+            raise ValueError('No indexes matching filters')
+        return indexes[0]
 
     def _attach_to_model(self, model):
         """
@@ -436,6 +466,14 @@ class RedisField(RedisProxyCommand):
             except AttributeError:
                 return True
 
+    @property
+    def attached_to_instance(self):
+        """Tells if the current field is the one attached to the instance"""
+        try:
+            return bool(self._instance)
+        except AttributeError:
+            return False
+
     def _call_command(self, name, *args, **kwargs):
         """
         Add lock management and call parent.
@@ -446,12 +484,14 @@ class RedisField(RedisProxyCommand):
                 try:
                     result = meth(name, *args, **kwargs)
                 except:
-                    self._rollback_indexes()
+                    if self._instance.connected:
+                        self._rollback_indexes()
                     raise
                 else:
                     return result
                 finally:
-                    self._reset_indexes_caches()
+                    if self._instance.connected:
+                        self._reset_indexes_rollback_caches(self._instance_pk)
         else:
             return meth(name, *args, **kwargs)
 
@@ -460,61 +500,77 @@ class RedisField(RedisProxyCommand):
         Restore the index in its previous status, using deindexed/indexed values
         temporarily stored.
         """
+        pk = self._instance.pk.get()
         for index in self._indexes:
-            index._rollback()
+            index._rollback(pk)
 
-    def _reset_indexes_caches(self):
+    def _reset_indexes_rollback_caches(self, pk):
         """
         Reset attributes used to store deindexed/indexed values, used to
         rollback the index when something failed.
         """
         for index in self._indexes:
-            index._reset_cache()
+            index._reset_rollback_cache(pk)
 
-    def index(self, value=None, only_index=None):
+    def get_for_instance(self, pk):
+        return self._model.lazy_connect(pk).get_field(self.name)
+
+    def _prepare_index_data(self, pk, values=None):
+        raise NotImplementedError
+
+    def _index(self, values, only_index=None):
         """
         Handle field index process.
         """
         assert self.indexable, "Field not indexable"
-        assert not only_index or self.has_index(only_index), "Invalid index"
         if only_index:
-            only_index = only_index if isclass(only_index) else only_index.__class__
+            indexes = [self.get_index(
+                index_class=only_index.__class__, key=only_index.key, prefix=only_index.prefix
+            )]
+        else:
+            indexes = self._indexes
 
-        if value is None:
-            value = self.proxy_get()
+        pk = self._instance.pk.get()
+        values = self._prepare_index_data(pk, values)
 
-        if value is not None:
-            needs_to_check_uniqueness = bool(self.unique)
+        for parts in values:
+            value = parts[-1]
+            if value is not None:
+                needs_to_check_uniqueness = bool(self.unique)
 
-            for index in self._indexes:
-                if only_index and not isinstance(index, only_index):
-                    continue
+                for index in indexes:
+                    index.add(
+                        pk,
+                        *parts,
+                        check_uniqueness=needs_to_check_uniqueness and index.handle_uniqueness
+                    )
 
-                index.add(value, check_uniqueness=needs_to_check_uniqueness and index.handle_uniqueness)
+                    if needs_to_check_uniqueness and index.handle_uniqueness:
+                        # uniqueness check is done for this value
+                        needs_to_check_uniqueness = False
 
-                if needs_to_check_uniqueness and index.handle_uniqueness:
-                    # uniqueness check is done for this value
-                    needs_to_check_uniqueness = False
-
-    def deindex(self, value=None, only_index=None):
+    def _deindex(self, values, only_index=None):
         """
         Run process of deindexing field value(s).
         """
         assert self.indexable, "Field not indexable"
-        assert not only_index or self.has_index(only_index), "Invalid index"
         if only_index:
-            only_index = only_index if isclass(only_index) else only_index.__class__
+            indexes = [self.get_index(
+                index_class=only_index.__class__, key=only_index.key, prefix=only_index.prefix
+            )]
+        else:
+            indexes = self._indexes
 
-        if value is None:
-            value = self.proxy_get()
+        pk = self._instance.pk.get()
+        values = self._prepare_index_data(pk, values)
 
-        if value is not None:
-            for index in self._indexes:
-                if only_index and not isinstance(index, only_index):
-                    continue
-                index.remove(value)
+        for parts in values:
+            value = parts[-1]
+            if value is not None:
+                for index in indexes:
+                    index.remove(pk, *parts)
 
-    def clear_indexes(self, chunk_size=1000, aggressive=False, index_class=None):
+    def clear_indexes(self, chunk_size=1000, aggressive=False):
         """Clear all indexes tied to this field
 
         Parameters
@@ -528,8 +584,6 @@ class RedisField(RedisProxyCommand):
             pattern of the keys used by the indexes. This is a lot faster and may find forgotten keys.
             But may also find keys not related to the index.
             Should be set to ``True`` if you are not sure about the already indexed values.
-        index_class: type
-            Allow to clear only index(es) for this index class instead of all indexes.
 
         Raises
         ------
@@ -541,7 +595,6 @@ class RedisField(RedisProxyCommand):
         --------
 
         >>> MyModel.get_field('myfield').clear_indexes()
-        >>> MyModel.get_field('myfield').clear_indexes(index_class=MyIndex)
 
         """
         assert self.indexable, "Field not indexable"
@@ -549,11 +602,9 @@ class RedisField(RedisProxyCommand):
             '`rebuild_indexes` can only be called on a field attached to the model'
 
         for index in self._indexes:
-            if index_class and not isinstance(index, index_class):
-                continue
             index.clear(chunk_size=chunk_size, aggressive=aggressive)
 
-    def rebuild_indexes(self, chunk_size=1000, aggressive_clear=False, index_class=None):
+    def rebuild_indexes(self, chunk_size=1000, aggressive_clear=False):
         """Rebuild all indexes tied to this field
 
         Parameters
@@ -564,8 +615,6 @@ class RedisField(RedisProxyCommand):
             Will be passed to the `aggressive` argument of the `clear_indexes` method.
             If `False`, all values will be normally deindexed. If `True`, the work
             will be done at low level, scanning for keys that may match the ones used by the indexes
-        index_class: type
-            Allow to build only index(es) for this index class instead of all indexes.
 
         Raises
         ------
@@ -577,7 +626,6 @@ class RedisField(RedisProxyCommand):
         --------
 
         >>> MyModel.get_field('myfield').rebuild_indexes()
-        >>> MyModel.get_field('myfield').clear_indexes(index_class=MyIndex)
 
 
         """
@@ -586,8 +634,6 @@ class RedisField(RedisProxyCommand):
             '`rebuild_indexes` can only be called on a field attached to the model'
 
         for index in self._indexes:
-            if index_class and not isinstance(index, index_class):
-                continue
             index.rebuild(chunk_size=chunk_size, aggressive_clear=aggressive_clear)
 
     def get_unique_index(self):
@@ -601,11 +647,18 @@ class RedisField(RedisProxyCommand):
                 (self._model.__name__, self.name)
             )
 
+    @property
+    def _instance_pk(self):
+        try:
+            return self._instance.pk.get()
+        except (ImplementationError, DoesNotExist):
+            return None
+
     def check_uniqueness(self, value):
         if not self.unique:
             return
         if value is not None:
-            self.get_unique_index().check_uniqueness(value)
+            self.get_unique_index().check_uniqueness(self._instance_pk, value)
 
     def from_python(self, value):
         """
@@ -637,6 +690,18 @@ class RedisField(RedisProxyCommand):
             self.index(result)
         return result
 
+    def _deny_if_indexable(self, command, *args, **kwargs):
+        """
+        Shortcut for commands that cannot be executed on indexable fields
+        """
+        if self.indexable:
+            raise ImplementationError('Indexable fields cannot be expired')
+        return self._traverse_command(command, *args, **kwargs)
+    _call_expire = _deny_if_indexable
+    _call_pexpire = _deny_if_indexable
+    _call_expireat = _deny_if_indexable
+    _call_pexpireat = _deny_if_indexable
+
     def _del(self, command, *args, **kwargs):
         """
         Shortcut for commands that remove all values of the field.
@@ -667,22 +732,41 @@ class SingleValueField(RedisField):
                     self.index(value)
         return self._traverse_command(command, value, *args, **kwargs)
 
+    def _prepare_index_data(self, pk, values=None):
+        if values is None:
+            values = [self.get_for_instance(pk).proxy_get()]
+        return [(value, ) for value in values]
+
+    def index(self, value=None, only_index=None):
+        self._index(None if value is None else [value], only_index)
+
+    def deindex(self, value=None, only_index=None):
+        self._deindex(None if value is None else [value], only_index)
+
 
 class StringField(SingleValueField):
 
     proxy_getter = "get"
     proxy_setter = "set"
 
-    available_getters = ('get', 'getbit', 'getrange', 'strlen', 'bitcount', )
-    available_modifiers = ('delete', 'getset', 'set', 'append', 'decr',
-                           'incr', 'incrbyfloat', 'setbit', 'setnx',
-                           'setrange', )
+    available_getters = SingleValueField.available_getters | {
+        'get', 'getbit', 'getrange', 'strlen', 'bitcount', 'bitpos',
+    }
+    available_modifiers = SingleValueField.available_modifiers | {
+        'delete', 'getset', 'set', 'append', 'decr', 'decrby',
+        'incr', 'incrby', 'incrbyfloat', 'setbit', 'setnx',
+        'setrange', 'setex', 'psetex',
+    }
 
     _call_getset = SingleValueField._call_set
     _call_append = _call_setrange = _call_setbit = SingleValueField._reset
     _call_decr = SingleValueField._reindex_from_result
+    _call_decrby = SingleValueField._reindex_from_result
     _call_incr = SingleValueField._reindex_from_result
+    _call_incrby = SingleValueField._reindex_from_result
     _call_incrbyfloat = SingleValueField._reindex_from_result
+    _call_setex = SingleValueField._deny_if_indexable
+    _call_psetex = SingleValueField._deny_if_indexable
 
     def _call_setnx(self, command, value):
         """
@@ -693,6 +777,19 @@ class StringField(SingleValueField):
             self.index(value)
         return result
 
+    def _call_set(self, command, value, ex=None, px=None, nx=False, xx=False):
+        """Deny expiring args if indexable, and deny other flags"""
+
+        if self.indexable and (ex is not None or px is not None):
+            raise ImplementationError('Indexable fields cannot be expired')
+
+        if nx:
+            raise LimpydException("nx argument to SET is not supported by limpyd")
+        if xx:
+            raise LimpydException("xx argument to SET is not supported by limpyd")
+
+        return super(StringField, self)._call_set(command, value, ex=ex, px=px)
+
 
 class MultiValuesField(RedisField):
     """
@@ -743,79 +840,36 @@ class MultiValuesField(RedisField):
             self.deindex([result])
         return result
 
+    def _prepare_index_data(self, pk, values=None):
+        if values is None:
+            values = self.get_for_instance(pk).proxy_get()
+        return [(value, ) for value in values]
+
     def index(self, values=None, only_index=None):
         """
         Index all values stored in the field, or only given ones if any.
         """
-        assert self.indexable, "Field not indexable"
-        assert not only_index or self.has_index(only_index), "Invalid index"
-        if only_index:
-            only_index = only_index if isclass(only_index) else only_index.__class__
-
-        if values is None:
-            values = self.proxy_get()
-
-        for value in values:
-            if value is not None:
-                needs_to_check_uniqueness = bool(self.unique)
-
-                for index in self._indexes:
-                    if only_index and not isinstance(index, only_index):
-                        continue
-
-                    index.add(value, check_uniqueness=needs_to_check_uniqueness and index.handle_uniqueness)
-
-                    if needs_to_check_uniqueness and index.handle_uniqueness:
-                        # uniqueness check is done for this value
-                        needs_to_check_uniqueness = False
+        self._index(values, only_index)
 
     def deindex(self, values=None, only_index=None):
         """
         Deindex all values stored in the field, or only given ones if any.
         """
-        assert self.indexable, "Field not indexable"
-        assert not only_index or self.has_index(only_index), "Invalid index"
-        if only_index:
-            only_index = only_index if isclass(only_index) else only_index.__class__
-
-        if not values:
-            values = self.proxy_get()
-
-        for value in values:
-            if value is not None:
-                for index in self._indexes:
-                    if only_index and not isinstance(index, only_index):
-                        continue
-                    index.remove(value)
+        self._deindex(values, only_index)
 
     def check_uniqueness(self, values):
         if not self.unique:
             return
+        pk = self._instance_pk
         for value in values:
             if value is not None:
-                self.get_unique_index().check_uniqueness(value)
+                self.get_unique_index().check_uniqueness(pk, value)
 
-    if redispy_version >= (2, 10, 0):
+    def _scan(self, command, match=None, count=None):
+        assert self.scannable, 'Field is not scannable'
+        assert not command.endswith('_iter'), 'Use %s instead of %s' % (command[:-5], command)
 
-        def _scan(self, command, match=None, count=None):
-            assert self.scannable, 'Field is not scannable'
-            assert not command.endswith('_iter'), 'Use %s instead of %s' % (command[:-5], command)
-
-            return self._traverse_command(command + '_iter', match=match, count=count)
-
-    else:
-        # no *scan_iter in redis-py < 2.10
-
-        def _scan(self, command, match=None, count=None):
-            assert self.scannable, 'Field is not scannable'
-
-            cursor = 0
-            while True:
-                cursor, data = self._traverse_command(command, cursor, match=match, count=count)
-                for item in (data.items() if hasattr(data, 'items') else data):
-                    yield item
-                if not cursor or cursor == '0':
-                    break
+        return self._traverse_command(command + '_iter', match=match, count=count)
 
 
 class SortedSetField(MultiValuesField):
@@ -831,20 +885,24 @@ class SortedSetField(MultiValuesField):
     proxy_getter = "zmembers"
     proxy_setter = "zadd"
 
-    available_getters = ('zcard', 'zcount', 'zrange', 'zrangebyscore',
-                         'zrank', 'zrevrange', 'zrevrangebyscore',
-                         'zrevrank', 'zscore', 'zscan', 'sort')
-    available_modifiers = ('delete', 'zadd', 'zincrby', 'zrem',
-                           'zremrangebyrank', 'zremrangebyscore', )
+    available_getters = MultiValuesField.available_getters | {
+        'zcard', 'zcount', 'zrange', 'zrangebyscore',
+        'zrank', 'zrevrange', 'zrevrangebyscore',
+        'zrevrank', 'zscore', 'zscan', 'sort', 'zscan_iter',
+        'zlexcount', 'zrangebylex', 'zrevrangebylex',
+    }
+    available_modifiers = MultiValuesField.available_modifiers | {
+        'delete', 'zadd', 'zincrby', 'zrem',
+        'zremrangebyrank', 'zremrangebyscore',
+        'zpopmin' ,'zpopmax', 'zremrangebylex',
+    }
 
     _call_zrem = MultiValuesField._rem
-    _call_zremrangebyscore = _call_zremrangebyrank = RedisField._reset
+    _call_zremrangebylex = _call_zremrangebyscore = _call_zremrangebyrank = RedisField._reset
 
     scannable = True
     _call_zscan = MultiValuesField._scan
-    if redispy_version >= (2, 10, 0):
-        _call_zscan_iter = MultiValuesField._scan
-        available_getters = available_getters + ('zscan_iter', )
+    _call_zscan_iter = MultiValuesField._scan
 
     def zmembers(self):
         """
@@ -854,67 +912,76 @@ class SortedSetField(MultiValuesField):
 
     def _call_zadd(self, command, *args, **kwargs):
         """
-        We do the same computation of the zadd method of StrictRedis to keep keys
-        to index them (instead of indexing the whole set)
-        Members (value/score) can be passed:
-            - in *args, with score followed by the value, 0+ times (to respect
-              the redis order)
-            - in **kwargs, with value as key and score as value
-        Example: zadd(1.1, 'my-key', 2.2, 'name1', 'name2', name3=3.3, name4=4.4)
+        Normal redis-py 3+ signature: mapping, nx=False, xx=False, ch=False, incr=False
+        So the normal way in redis-py 3+ to pass data to zadd is to pass a mapping.
+        But to avoid too much rewriting when using the version of limpyd supporting
+        redis-py 3+, we still allow passing data as named arguments.
+        The old way of passing scores and values as unnamed argument is not supported anymore.
+        When passing data as named arguments, special redis arguments, ie nx, xx, ch and incr
+        will NOT be treated as data.
+        We actually don't support xx, nx and incr options.
         """
+        args, kwargs = self.coerce_zadd_args(*args, **kwargs)
         if self.indexable:
-            keys = []
-            if args:
-                if len(args) % 2 != 0:
-                    raise RedisError("ZADD requires an equal number of "
-                                     "values and scores")
-                keys.extend(args[1::2])
-            keys.extend(kwargs)  # add kwargs keys (values to index)
-            self.index(keys)
+            mapping = args[0] if args else kwargs['mapping']
+            self.index(mapping.keys())
         return self._traverse_command(command, *args, **kwargs)
 
-    def _call_zincrby(self, command, value, *args, **kwargs):
+    def _call_zincrby(self, command, amount, value):
         """
         This command update a score of a given value. But it can be a new value
         of the sorted set, so we index it.
         """
         if self.indexable:
             self.index([value])
-        return self._traverse_command(command, value, *args, **kwargs)
+        return self._traverse_command(command, amount, value)
 
     @staticmethod
     def coerce_zadd_args(*args, **kwargs):
         """
-        Take arguments attended by a zadd call, named or not, and return a flat list
-        that can be used.
+        Take arguments attended by a zadd call, named or not, and return kwargs that can be used
+        directly..
         A callback can be called with all "values" (as *args) if defined as the
         `values_callback` named argument. Real values will then be the result of
         this callback.
         """
         values_callback = kwargs.pop('values_callback', None)
 
-        pieces = []
+        # if args, it must be two max, with first being the mapping and if two, the second is `ch`
         if args:
-            if len(args) % 2 != 0:
-                raise RedisError("ZADD requires an equal number of "
-                                 "values and scores")
-            pieces.extend(args)
+            if not kwargs and len(args) > 2 or kwargs and (len(args) > 1 or list(kwargs.keys()) != ['ch']):
+                raise LimpydException("ZADD accepts only two argumnets: 'mapping' and 'ch'")
+            kwargs['mapping'] = args[0]
+            if len(args) > 1:
+                kwargs['ch'] = args[1]
+            args = []
 
-        for pair in iteritems(kwargs):
-            pieces.append(pair[1])
-            pieces.append(pair[0])
+        if 'mapping' not in kwargs:
+            # handle our special case allowing passing value=score arguments
+            mapping = kwargs.copy()
+            final_kwargs = {}
+            for key in {'nx', 'xx', 'ch', 'incr'}:
+                if key in kwargs:
+                    final_kwargs[key] = mapping.pop(key)
+            kwargs = final_kwargs
+            kwargs['mapping'] = mapping
+
+        for key in {'nx', 'xx', 'incr'}:
+            if key in kwargs:
+                raise LimpydException("%s argument to ZADD is not supported by limpyd" % key)
 
-        values = pieces[1::2]
         if values_callback:
-            values = values_callback(*values)
+            mapping = args[0] if args else kwargs['mapping']
+            for oldkey, newkey in list(zip(mapping.keys(), values_callback(*mapping.keys()))):
+                mapping[newkey] = mapping.pop(oldkey)
 
-        scores = pieces[0::2]
+        return args, kwargs
 
-        pieces = []
-        for z in zip(scores, values):
-            pieces.extend(z)
-
-        return pieces
+    def _call_zpopmax(self, command, *args, **kwargs):
+        if self.database.redis_version < (5, ):
+            raise ImplementationError("%s is not a valid command for redis-server version < 5" % command.upper())
+        return self._reset(command, *args, **kwargs)
+    _call_zpopmin = _call_zpopmax
 
 
 class SetField(MultiValuesField):
@@ -928,18 +995,31 @@ class SetField(MultiValuesField):
     proxy_getter = "smembers"
     proxy_setter = "sadd"
 
-    available_getters = ('scard', 'sismember', 'smembers', 'srandmember', 'sscan', 'sort', )
-    available_modifiers = ('delete', 'sadd', 'srem', 'spop', )
+    available_getters = MultiValuesField.available_getters | {
+        'scard', 'sismember', 'smembers', 'srandmember', 'sscan', 'sort', 'sscan_iter',
+    }
+    available_modifiers = MultiValuesField.available_modifiers | {
+        'delete', 'sadd', 'srem', 'spop',
+    }
 
     _call_sadd = MultiValuesField._add
     _call_srem = MultiValuesField._rem
-    _call_spop = MultiValuesField._pop
 
     scannable = True
     _call_sscan = MultiValuesField._scan
-    if redispy_version >= (2, 10, 0):
-        _call_sscan_iter = MultiValuesField._scan
-        available_getters = available_getters + ('sscan_iter', )
+    _call_sscan_iter = MultiValuesField._scan
+
+    def _call_spop(self, command, count=None):
+        if count is not None and self.database.redis_version < (3, 2):
+            raise ImplementationError("Count argument to SPOP is invalid for redis-server version < 3.2")
+
+        if count and self.indexable:
+            # deindex all returned values (`_pop` assuming only one value)
+            result = self._traverse_command(command, count=count)
+            self.deindex(result)
+            return result
+
+        return super(SetField, self)._pop(command, count=count)
 
 
 class ListField(MultiValuesField):
@@ -957,21 +1037,93 @@ class ListField(MultiValuesField):
     proxy_getter = "lmembers"
     proxy_setter = "rpush"
 
-    available_getters = ('lindex', 'llen', 'lrange', )
-    available_modifiers = ('delete', 'linsert', 'lpop', 'lpush', 'lpushx',
-                           'lrem', 'rpop', 'rpush', 'rpushx', 'lset',
-                           'ltrim', 'sort', )
+    available_getters = MultiValuesField.available_getters | {
+        'lindex', 'llen', 'lrange', 'lrank', 'lcontains', 'lcount',
+    }
+    available_modifiers = MultiValuesField.available_modifiers | {
+        'delete', 'linsert', 'lpop', 'lpush', 'lpushx',
+        'lrem', 'rpop', 'rpush', 'rpushx', 'lset',
+        'ltrim', 'sort',
+    }
 
     _call_lpop = _call_rpop = MultiValuesField._pop
     _call_lpush = _call_rpush = MultiValuesField._add
     _call_ltrim = RedisField._reset
 
+    scripts = {
+        'lrank': {
+            # get position of a value in a list
+            'lua': """
+                local list_key = KEYS[1]
+                local value = ARGV[1]
+                local items = redis.call('lrange', list_key, 0, -1)
+                for i, item in ipairs(items) do
+                    if items[i] == value then
+                        return i - 1
+                    end
+                end
+                return nil
+            """,
+        },
+        'lcount': {
+            # count occurrences of a value in a list
+            'lua': """
+                local list_key = KEYS[1]
+                local value = ARGV[1]
+                local items = redis.call('lrange', list_key, 0, -1)
+                local count = 0
+                for i, item in ipairs(items) do
+                    if items[i] == value then
+                        count = count + 1
+                    end
+                end
+                return count
+            """,
+        },
+    }
+
     def lmembers(self):
         """
         Used as a proxy_getter to get all values stored in the field.
         """
         return self.lrange(0, -1)
 
+    def _call_lrank(self, command, value):
+        """
+        Addon to redis, to know if a value is in the list without having to retrieve all the list,
+        thanks to lua scripting.
+        Returns None if the value is not in the list, else its position, 0-indexed.
+        """
+        return self.database.call_script(
+            # be sure to use the script dict at the class level
+            # to avoid registering it many times
+            script_dict=self.__class__.scripts[command],
+            keys=[self.key],
+            args=[value]
+        )
+
+    def _call_lcontains(self, command, value):
+        """
+        Addon to redis, to know if a value is in the list without having to retrieve all the list,
+        thanks to lua scripting. This is done via lrank defined above
+        Returns a boolean indicating if the value is in the list.
+        """
+        return self.lrank(value) is not None
+
+    def _call_lcount(self, command, value):
+        """
+        Addon to redis, to know how many times value is in the list without having to retrieve all
+        the list, thanks to lua scripting.
+        Returns an integer: 0 if not found, else the number of occurences
+        """
+        return self.database.call_script(
+            # be sure to use the script dict at the class level
+            # to avoid registering it many times
+            script_dict=self.__class__.scripts[command],
+            keys=[self.key],
+            args=[value]
+        )
+
     def _pushx(self, command, *args, **kwargs):
         """
         Helper for lpushx and rpushx, that only index the new values if the list
@@ -1026,22 +1178,24 @@ class HashField(MultiValuesField):
     proxy_getter = "hgetall"
     proxy_setter = "hmset"
 
-    available_getters = ('hget', 'hgetall', 'hmget', 'hkeys', 'hvals',
-                         'hlen', 'hscan', )
-    available_modifiers = ('delete', 'hdel', 'hmset', 'hsetnx', 'hset',
-                           'hincrby', 'hincrbyfloat', )
+    available_getters = MultiValuesField.available_getters | {
+        'hget', 'hgetall', 'hmget', 'hkeys', 'hvals',
+        'hlen', 'hscan', 'hscan_iter', 'hstrlen',
+    }
+    available_modifiers = MultiValuesField.available_modifiers | {
+        'delete', 'hdel', 'hmset', 'hsetnx', 'hset',
+        'hincrby', 'hincrbyfloat',
+    }
 
     scannable = True
     _call_hscan = MultiValuesField._scan
-    if redispy_version >= (2, 10, 0):
-        _call_hscan_iter = MultiValuesField._scan
-        available_getters = available_getters + ('hscan_iter', )
+    _call_hscan_iter = MultiValuesField._scan
 
     def _call_hmset(self, command, *args, **kwargs):
         if self.indexable:
-            current = self.proxy_get()
-            _to_deindex = dict((k, current[k]) for k in iterkeys(kwargs) if k in current)
-            self.deindex(_to_deindex)
+            keys = list(kwargs.keys())
+            current = self.hmget(*keys)
+            self.deindex({key: value for key, value in zip(keys, current) if value is not None})
             self.index(kwargs)
         return self._traverse_command(command, kwargs)
 
@@ -1068,8 +1222,8 @@ class HashField(MultiValuesField):
 
     def _call_hdel(self, command, *args):
         if self.indexable:
-            current = self.proxy_get()
-            self.deindex(dict((k, current[k]) for k in args if k in current))
+            current = self.hmget(*args)
+            self.deindex({key: value for key, value in zip(args, current) if value is not None})
         return self._traverse_command(command, *args)
 
     def _call_hsetnx(self, command, key, value):
@@ -1083,49 +1237,27 @@ class HashField(MultiValuesField):
         # redispy needs a list, not args
         return self._traverse_command(command, args)
 
+    def _call_hstrlen(self, command, key):
+        if self.database.redis_version < (3, 2):
+            raise ImplementationError("HSTRLEN is not a valid command for redis-server version < 3.2")
+        return self._traverse_command(command, key)
+
+    def _prepare_index_data(self, pk, values=None):
+        if values is None:
+            values = self.get_for_instance(pk).proxy_get()
+        return list(iteritems(values))
+
     def index(self, values=None, only_index=None):
         """
         Deal with dicts and field names.
         """
-        assert self.indexable, "Field not indexable"
-        assert not only_index or self.has_index(only_index), "Invalid index"
-        if only_index:
-            only_index = only_index if isclass(only_index) else only_index.__class__
-
-        if values is None:
-            values = self.proxy_get()
-
-        for field_name, value in iteritems(values):
-            if value is not None:
-                needs_to_check_uniqueness = bool(self.unique)
-
-                for index in self._indexes:
-                    if only_index and not isinstance(index, only_index):
-                        continue
-
-                    index.add(field_name, value, check_uniqueness=needs_to_check_uniqueness and index.handle_uniqueness)
-
-                    if needs_to_check_uniqueness and index.handle_uniqueness:
-                        # uniqueness check is done for this value
-                        needs_to_check_uniqueness = False
+        self._index(values, only_index)
 
     def deindex(self, values=None, only_index=None):
         """
         Deal with dicts and field names.
         """
-        assert self.indexable, "Field not indexable"
-        assert not only_index or self.has_index(only_index), "Invalid index"
-        if only_index:
-            only_index = only_index if isclass(only_index) else only_index.__class__
-
-        if values is None:
-            values = self.proxy_get()
-        for field_name, value in iteritems(values):
-            if value is not None:
-                for index in self._indexes:
-                    if only_index and not isinstance(index, only_index):
-                        continue
-                    index.remove(field_name, value)
+        self._deindex(values, only_index)
 
     def hexists(self, key):
         """
@@ -1152,9 +1284,8 @@ class InstanceHashField(SingleValueField):
     proxy_getter = "hget"
     proxy_setter = "hset"
 
-    available_getters = ('hget', )
-    available_modifiers = ('hdel', 'hset', 'hsetnx', 'hincrby',
-                           'hincrbyfloat', )
+    available_getters = {'hget', }
+    available_modifiers = {'hdel', 'hset', 'hsetnx', 'hincrby', 'hincrbyfloat', }
 
     _call_hset = SingleValueField._call_set
     _call_hdel = RedisField._del
@@ -1211,8 +1342,8 @@ class PKField(SingleValueField):
     proxy_getter = "get"
     proxy_setter = "set"
 
-    available_getters = ('get',)
-    available_modifiers = ('set',)
+    available_getters = {'get', }
+    available_modifiers = {'set', }
 
     name = 'pk'  # Default name ok the pk, can be changed by declaring a new PKField
     indexable = False  # Not an `indexable` field...
@@ -1300,7 +1431,7 @@ class PKField(SingleValueField):
         self._set = True
 
         # We have a new pk, so add it to the collection
-        log.debug("Adding %s in %s collection" % (value, self._model._name))
+        log.debug("Adding %s in %s collection" % (value, self._model.__name__))
         self.connection.sadd(self.collection_key, value)
 
         # Finally return 1 as we did a real redis call to the set command
@@ -1314,6 +1445,7 @@ class PKField(SingleValueField):
         if not hasattr(self, '_instance'):
             raise ImplementationError("Impossible to get the PK of an unbound field")
         if not hasattr(self._instance, '_pk'):
+            self._instance._connected = False
             raise DoesNotExist("The current object doesn't exists anymore")
 
         if not self._instance._pk:
@@ -1356,7 +1488,8 @@ class FieldLock(Lock):
     release.
     """
 
-    def __init__(self, field, timeout=5, sleep=0.1):
+    def __init__(self, field, timeout=5, sleep=0.1,
+                 blocking=True, blocking_timeout=None, thread_local=True):
         """
         Save the field and create a real lock,, using the correct connection
         and a computed lock key based on the names of the field and its model.
@@ -1368,6 +1501,9 @@ class FieldLock(Lock):
             name=make_key(field._model._name, 'lock-for-update', field.name),
             timeout=timeout,
             sleep=sleep,
+            blocking=blocking,
+            blocking_timeout=blocking_timeout,
+            thread_local=thread_local
         )
 
     def _get_already_locked_by_model(self):
@@ -1391,12 +1527,12 @@ class FieldLock(Lock):
         sub-lock status.
         """
         if not self.field.lockable:
-            return
+            return True
         if self.already_locked_by_model:
             self.sub_lock_mode = True
-            return
+            return True
         self.already_locked_by_model = True
-        super(FieldLock, self).acquire(*args, **kwargs)
+        return super(FieldLock, self).acquire(*args, **kwargs)
 
     def release(self, *args, **kwargs):
         """
